/**
 * My Nixie Clock IN-12
 * Vladimir N. Shilov <shilow@ukr.net>
 * 2019.01.18
 */

#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/eeprom.h>

#include "main.h"
#include "i2c.h"
#include "ds3231.h"
#include "rtos.h"
#include "event-system.h"

/* Defines */
#define TIMER2_HZ         250
#define TIMER2_PRESCALER  256
#define TIMER2_CS         (1<<CS22 | 1<<CS21 | 0<<CS20)
#define TIMER2_CNT        (0x100 - (F_CPU / TIMER2_PRESCALER / TIMER2_HZ))
#define OCR2_MIN          (0xFF - TIMER2_CNT)
#define OCR2_MAX          0xFF
#define FULL_BRIGHT_ON    0x06
#define FULL_BRIGHT_OFF   0x22

/* Variables */
static volatile uint8_t Digit[4] = {1, 2, 3, 4};
static rtc_t RTC, setRTC;
static volatile struct {
  uint8_t RTC_Int:  1;
  uint8_t saveCal:  1;
  uint8_t blinkC:   1;
  uint8_t blink0:   1;
  uint8_t blink1:   1;
  uint8_t blink2:   1;
  uint8_t blink3:   1;
  uint8_t saveEEP:  1;
} Flag;
static btn_t Button[BTN_NUM] = {
  {0, evBTN1Pressed, evBTN1Holded, BUTTON1_PIN},
  {0, evBTN2Pressed, evBTN2Pressed, BUTTON2_PIN},
  {0, evBTN3Pressed, evBTN2Pressed, BUTTON3_PIN}
};
static volatile uint8_t DISP_WDT = 0;
static EEMEM uint8_t EEP_Bright;
static uint8_t Bright;
static EEMEM uint8_t EEP_SunTime;

/* Constants */
/* Function prototypes */
static void Board_Init(void);
static void dotOn(void);
static void btnProcess(void);
static void valIncrease(uint8_t * val, uint8_t max);
static void valDecrease(uint8_t * val, uint8_t max);
static void blink(void);
static void setSunWinterTime(void);

void main(void) {
  /**
   * Локальные переменные
   */
  uint8_t event = 0;
  Flag.RTC_Int = 0;
  Flag.saveCal = 0;
  Flag.blink0 = 0;
  Flag.blink1 = 0;
  Flag.blink2 = 0;
  Flag.blink3 = 0;
  Flag.blinkC = 0;
  Flag.saveEEP = 0;

  Bright = eeprom_read_byte(&EEP_Bright);

  /**
   * Инициализация, настройка...
   */
  Board_Init();

  /* Initialize Scheduler */
  RTOS_Init();

  /* Initialize I2C Bus and RTC */
  I2C_Init();
  RTC_Init();
  RTC_ReadAll(&RTC);

  /* Initialize Event State Machine */
  ES_Init(stShowTime);

  showTime();

  RTOS_SetTask(btnProcess, 5, BTN_SCAN_PERIOD);

  /** main loop */
  do {
    if (Flag.RTC_Int != 0) {
      Flag.RTC_Int = 0;
      ES_PlaceEvent(evNewSecond);

      RTC_ReadTime(&RTC);
      if (RTC.Sec == 0 && RTC.Min == 0) {
        // begin of new hour
        if (RTC.Hr == 0) {
          RTC_ReadCalendar(&RTC);
          ES_PlaceEvent(evRefreshCal);
        }
        if (RTC.Hr >= FULL_BRIGHT_ON && RTC.Hr < FULL_BRIGHT_OFF) {
          OCR2 = OCR2_MAX;
        } else {
          OCR2 = Bright;
        }

        setSunWinterTime();
      } // begin new hour

      if (DISP_WDT != 0) {
        DISP_WDT --;
        if (DISP_WDT == 0) {
          ES_PlaceEvent(evDisplayWDT);

          if (Flag.saveCal != 0) {
            Flag.saveCal = 0;
            RTC_WriteCalendar(&RTC);
          }

          if (Flag.saveEEP != 0) {
            Flag.saveEEP = 0;
            eeprom_update_byte(&EEP_Bright, Bright);
          }
        }
      }
    }

    event = ES_GetEvent();
    if (event) {
      ES_Dispatch(event);
    }

    // крутим диспетчер
    RTOS_DispatchTask();

    // делать нечего -- спим, ждём прерывание
    set_sleep_mode(SLEEP_MODE_IDLE);
    sleep_mode();

  } while(1);
}

/**
 *  П о д п р о г р а м м ы
 */

/**
 * @brief Initializy perephireal
 */
static void Board_Init(void) {
  /* power off Analog Comparator */
  ACSR = ACD;

  /* GPIO */
  PORTB = BUTTON_PINS; // enable pull-up
  DDRC = DIGIT_PINS; // as output
  DDRD = (DOT_PIN | ANOD_PINS); // as output

  /* Timer2 - refresh Nixie values */
  TCCR2 = TIMER2_CS;
  TCNT2 = TIMER2_CNT;
  TIMSK = _BV(TOIE2) | _BV(OCIE2);
  OCR2 = Bright;

  /* Interrupt from RTC */
  MCUCR = _BV(ISC11); // falling edge
  GICR = _BV(INT1);

  /* Enable Interrupts */
  sei();
}

/**
 * @brief Correct current time for Sun or Winter
 */
static void setSunWinterTime(void) {
  uint8_t sunTime = eeprom_read_byte(&EEP_SunTime);

  /* Переход на летнее время */
  if ((RTC.Mon == 3) && (RTC.WD == 7) && (RTC.Hr == 3) && (sunTime != 0)) {
    if ((RTC.Day + 7) > 31) {
      RTC.Hr = 4;
      RTC_WriteHHMM(&RTC);
      sunTime = 0;
    }
  }

  /* Переход на зимнее время */
  if ((RTC.Mon == 10) && (RTC.WD == 7) && (RTC.Hr == 4) && (sunTime == 0)) {
    if ((RTC.Day + 7) > 31) {
      RTC.Hr = 3;
      RTC_WriteHHMM(&RTC);
      sunTime = 1;
    }
  }

  eeprom_update_byte(&EEP_SunTime, sunTime);
}

static void dotOn(void) {
  PORTD |= DOT_PIN;
}

void dotOff(void) {
  PORTD &= ~(DOT_PIN);
}

void dotOnPersistent(void) {
  RTOS_DeleteTask(dotOff);
  PORTD |= DOT_PIN;
}

/**
  * @brief  Обработка кнопок.
  * @param  : None
  * @retval : None
  */
static void btnProcess(void) {
  uint8_t i;
  for (i=0; i<BTN_NUM; i++) {
    if (Button[i].pin != 0) {
      // button pressed
      if (BUTTON_STATE(Button[i].pin) == 0) {
        Button[i].time ++;
        if (Button[i].time > BTN_TIME_HOLDED) {
          Button[i].time -= BTN_TIME_REPEATED;
          if (Button[i].holded == Button[i].pressed) {
            // if pressed and holded - same function, then button pressed auto repeat
            ES_PlaceEvent(Button[i].pressed);
          }
        }
      } else {
        // button released
        if (Button[i].time >= (BTN_TIME_HOLDED - BTN_TIME_REPEATED)) {
          ES_PlaceEvent(Button[i].holded); // process long press
        } else if (Button[i].time >= BTN_TIME_PRESSED) {
          ES_PlaceEvent(Button[i].pressed); // process short press
        }
        Button[i].time = 0;
        RTOS_SetTask(btnProcess, BTN_TIME_PAUSE, BTN_SCAN_PERIOD);
      }
    } /* end (pin == 0) */
  } /* end FOR */
}

void showTime(void) {
  dotOn();
  RTOS_SetTask(dotOff, 500, 0);

  if (RTC.Hr > 0x09) {
    Digit[0] = RTC.Hr >> 4;
  } else {
    Digit[0] = DIGIT_BLANK;
  }
  Digit[1] = RTC.Hr & 0x0F;
  Digit[2] = RTC.Min >> 4;
  Digit[3] = RTC.Min & 0x0F;
}

void showMMSS(void) {
  Digit[0] = RTC.Min >> 4;
  Digit[1] = RTC.Min & 0x0F;
  Digit[2] = RTC.Sec >> 4;
  Digit[3] = RTC.Sec & 0x0F;
}

void showWDay(void) {
  DISP_WDT = DISP_WDT_TIME;
  Digit[0] = DIGIT_BLANK;
  Digit[1] = DIGIT_BLANK;
  Digit[2] = RTC.WD & 0x0F;
  Digit[3] = DIGIT_BLANK;
}

void showMDay(void) {
  DISP_WDT = DISP_WDT_TIME;
  Digit[0] = RTC.Day >> 4;
  Digit[1] = RTC.Day & 0x0F;
  Digit[2] = DIGIT_BLANK;
  Digit[3] = DIGIT_BLANK;
}

void showMonth(void) {
  DISP_WDT = DISP_WDT_TIME;
  Digit[0] = DIGIT_BLANK;
  Digit[1] = DIGIT_BLANK;
  Digit[2] = RTC.Mon >> 4;
  Digit[3] = RTC.Mon & 0x0F;
}

void showYear(void) {
  DISP_WDT = DISP_WDT_TIME;
  Digit[0] = 0x02;
  Digit[1] = 0x00;
  Digit[2] = RTC.Year >> 4;
  Digit[3] = RTC.Year & 0x0F;
}

void showBright(void) {
  DISP_WDT = DISP_WDT_TIME;
  Digit[0] = DIGIT_BLANK;
  Digit[1] = 0x00;
  Digit[2] = RTC.Year >> 4;
  Digit[3] = RTC.Year & 0x0F;
}

void incWDay(void) {
  if (RTC.WD < 7) {
    RTC.WD ++;
  } else {
    RTC.WD = 1;
  }
  Flag.saveCal = 1;
}

void decWDay(void) {
  if (RTC.WD > 1) {
    RTC.WD --;
  } else {
    RTC.WD = 7;
  }
  Flag.saveCal = 1;
}

void incMDay(void) {
  valIncrease(&RTC.Day, 31);
  Flag.saveCal = 1;
}

void decMDay(void) {
  valDecrease(&RTC.Day, 31);
  Flag.saveCal = 1;
}

void incMonth(void) {
  valIncrease(&RTC.Mon, 12);
  if (RTC.Mon == 0) {
    RTC.Mon = 1;
  }
  Flag.saveCal = 1;
}

void decMonth(void) {
  valDecrease(&RTC.Mon, 12);
  if (RTC.Mon == 0) {
    RTC.Mon = 0x12;
  }
  Flag.saveCal = 1;
}

void incYear(void) {
  valIncrease(&RTC.Year, 99);
  Flag.saveCal = 1;
}

void decYear(void) {
  valDecrease(&RTC.Year, 99);
  Flag.saveCal = 1;
}

void incBright(void) {
  if (Bright < 0xFF) {
    Bright ++;
  } else {
    Bright = OCR2_MIN;
  }
  OCR2 = Bright;
  Flag.saveEEP = 1;
}

void decBright(void) {
  if (Bright > OCR2_MIN) {
    Bright --;
  } else {
    Bright = 0xFF;
  }
  OCR2 = Bright;
  Flag.saveEEP = 1;
}

static void blink(void) {
  static uint8_t s = 0;
  switch (s) {
  case 0:
    Flag.blinkC = 0;
    RTOS_SetTask(blink, 750, 0);
    s = 1;
    break;

  case 1:
    Flag.blinkC = 1;
    RTOS_SetTask(blink, 250, 0);
    s = 0;
    break;

  default:
    s = 0;
  }
}

void setTimeShow(void) {
  dotOn();
  RTOS_SetTask(dotOff, 500, 0);

  Digit[0] = setRTC.Hr >> 4;
  Digit[1] = setRTC.Hr & 0x0F;
  Digit[2] = setRTC.Min >> 4;
  Digit[3] = setRTC.Min & 0x0F;
}

void setTimeBegin(void) {
  RTC_ReadTime(&setRTC);
  RTOS_SetTask(btnProcess, 500, BTN_SCAN_PERIOD);
}

void setHHBegin(void) {
  Flag.blink0 = 1;
  Flag.blink1 = 1;
  Flag.blink2 = 0;
  Flag.blink3 = 0;
  RTOS_SetTask(blink, 0, 0);
  setTimeShow();
}

void setHHInc(void) {
  valIncrease(&setRTC.Hr, 23);
}

void setHHDec(void) {
  valDecrease(&setRTC.Hr, 23);
}

void setMMBegin(void) {
  Flag.blink0 = 0;
  Flag.blink1 = 0;
  Flag.blink2 = 1;
  Flag.blink3 = 1;
  RTOS_SetTask(blink, 0, 0);
  setTimeShow();
}

void setMMInc(void) {
  valIncrease(&setRTC.Min, 59);
}

void setMMDec(void) {
  valDecrease(&setRTC.Min, 59);
}

void setTimeEnd(void) {
  RTOS_SetTask(btnProcess, 500, BTN_SCAN_PERIOD);

  setRTC.Sec = 0;
  RTC_WriteTime(&setRTC);

  RTOS_DeleteTask(blink);
  Flag.blink0 = 0;
  Flag.blink1 = 0;
  Flag.blink2 = 0;
  Flag.blink3 = 0;
  Flag.blinkC = 0;

  RTC_ReadTime(&RTC);
}

/**
  * @brief  Increase BCD value.
  * @param  : val, max
  * @retval : None
  */
static void valIncrease(uint8_t * val, uint8_t max) {
  uint8_t bin = 10 * (*val >> 4) + (*val & 0x0f);
  if (bin < max) {
    bin ++;
  } else {
    bin = 0;
  }
  *val = ((bin / 10 ) << 4) | (bin % 10);
}

/**
  * @brief  Decrease BCD value.
  * @param  : value, max
  * @retval : None
  */
static void valDecrease(uint8_t * val, uint8_t max) {
  uint8_t bin = 10 * (*val >> 4) + (*val & 0x0f);
  if (bin > 0) {
    bin --;
  } else {
    bin = max;
  }
  *val = ((bin / 10 ) << 4) | (bin % 10);
}


/**
 *  П р е р ы в а н и я
 */

/**
 * @brief RTC one seconds interrupt
 */
ISR (INT1_vect) {
  Flag.RTC_Int = 1;
}

/**
 * @brief Refresh Nixie output
 * @note  Digit[] must be in range 0x00 - 0x0F
 */
ISR(TIMER2_OVF_vect) {
  static uint8_t idx = 0;

  // reload timer
  TCNT2 = TIMER2_CNT;

  // read current register value and clean bits
  uint8_t pd = PORTD & ~ANOD_PINS;
  uint8_t pc = PORTC & ~DIGIT_PINS;

  // power off lamps
//  PORTD = pd;
//  PORTC = pc;

  switch (idx) {
  case 0:
    // output lamp value
    PORTC = pc | Digit[0];
    // power on lamp
    if (Digit[0] != DIGIT_BLANK) {
      if (Flag.blink0 == 0 || Flag.blinkC == 0) {
        PORTD = pd | ANOD1;
      }
    }
    idx = 1;
    break;

  case 1:
    PORTC = pc | Digit[1];
    if (Digit[1] != DIGIT_BLANK) {
      if (Flag.blink1 == 0 || Flag.blinkC == 0) {
        PORTD = pd | ANOD2;
      }
    }
    idx = 2;
    break;

  case 2:
    PORTC = pc | Digit[2];
    if (Digit[2] != DIGIT_BLANK) {
      if (Flag.blink2 == 0 || Flag.blinkC == 0) {
        PORTD = pd | ANOD3;
      }
    }
    idx = 3;
    break;

  case 3:
    PORTC = pc | Digit[3];
    if (Digit[3] != DIGIT_BLANK) {
      if (Flag.blink3 == 0 || Flag.blinkC == 0) {
        PORTD = pd | ANOD4;
      }
    }
    idx = 0;
    break;

  default:
    idx = 0;
    break;
  }
}

/**
 * @brief Power Off Nixie output
 * @note  For Brightnes dimming
 */
ISR(TIMER2_COMP_vect) {
  // power off lamps
  PORTD &= ~ANOD_PINS;
  PORTC &= ~DIGIT_PINS;
}

/**
 * @brief заглушка для неиспользуемых прерываний
 */
ISR(__vector_default,ISR_NAKED) {
  reti();
}
